import argparse
import gc
import importlib.resources as resources
import logging
import os
import random
import re
import shlex
import socket
import subprocess
import threading
import time
import xml.etree.ElementTree as ET
from datetime import datetime
from typing import List, Optional, Tuple, Union

import psutil
import pynvml
import torch
from prompt_toolkit import PromptSession, prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.shortcuts import print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator
from spellchecker import SpellChecker
from termcolor import colored, cprint
from tqdm import tqdm
from transformers import GPT2LMHeadModel, GPT2Tokenizer
from transformers import logging as trans_log
from whoosh.analysis import StandardAnalyzer
from whoosh.fields import ID, TEXT, Schema
from whoosh.index import open_dir
from whoosh.qparser import MultifieldParser, OrGroup

trans_log.set_verbosity_error()
analyzer = StandardAnalyzer(stoplist=None)
schema = Schema(
    title=TEXT(stored=True, analyzer=analyzer),
    path=ID(stored=True),
    content=TEXT(stored=True, analyzer=analyzer),
)
logging.basicConfig(
    filename="command_errors.log",
    level=logging.ERROR,
    format="%(asctime)s - %(levelname)s - %(message)s",
)
MAX_RESULTS_DEFAULT = 1e6


class ActionChoiceValidator(Validator):
    def validate(self, document):
        text = document.text.lower().strip()
        if text not in ["yes", "y", "no", "n", "always", "a"]:
            raise ValidationError(
                message="Invalid choice. Please enter 'y', 'n', or 'a'."
            )


class FunctionValidator(Validator):
    def __init__(self, validation_function):
        self.validation_function = validation_function

    def validate(self, document):
        if not self.validation_function(document.text):
            raise ValidationError(message="Invalid input. Please try again.")


class WordValidator(Validator):
    def __init__(self, accepted_words):
        self.accepted_words = accepted_words

    def validate(self, document):
        if document.text.lower() not in self.accepted_words:
            raise ValidationError(message="Please enter a valid choice.")


class InteractiveGenerator:
    IP_PATTERN = re.compile(r"\b\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(?:/\d{1,2})?\b")
    FLAG_PATTERN = re.compile(r"(-\w+|--[\w-]+)")  # Updated Regular expression
    URL_PATTERN_VALIDATION = r"http[s]?://(?:[a-zA-Z]|[0-9]|[-._~:/?#[\]@!$&'()*+,;=]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
    CVE_PATTERN = re.compile(
        r"CVE-\d{4}-\d{4,7}", re.IGNORECASE
    )  # Regular expression for CVE pattern

    def __init__(self, results_dir=None, model_dir=None, testing_mode=None):
        self.args = self._parse_arguments()
        if results_dir is not None:
            self.args.results_dir = results_dir
        if model_dir is not None:
            self.args.model_dir = model_dir
        if testing_mode is not None:
            self.args.testing_mode = testing_mode
        self.index_dir = self.return_path("indexdir")
        self.s3_url = self._determine_s3_url()
        self._ensure_model_folder_exists()
        self._validate_model_dirs()
        self.command_running = False
        self._ensure_results_directory_exists()
        self.max_truncate_length: int = 500
        self.single_model_mode = False
        self.tokenizers = {}
        self.models = {}
        self.print_star_sky()
        self.random_name = None
        self.current_model = None
        self.current_model_name = None
        self.print_lock = threading.Lock()
        self.model_names = self.get_model_names()
        self.always_apply_action: bool = False
        if self.args.autonomous_mode is True:
            self.single_model_mode = True
            self.autonomous_mode()
        self.args.autonomous_mode = False
        if self.args.testing_mode:
            cprint("testing completed, exiting...", "green")
            exit()
        while True:  # This loop will keep asking until a valid mode is selected
            mode = self.select_mode()
            if mode == "a":
                self.current_model_name = self._load_tokenizer_and_model()
                break
            elif mode == "d":
                self.single_model_mode = True
                self.current_model_name = self._select_model()
                break
            else:
                cprint("\nInvalid Choice", "red")

        self.log_file_path = None
        self.services = []
        self.flag_file = self.return_path(self.current_model_name + "_flags")
        self.flag_descriptions = self._load_flag_descriptions(self.flag_file)
        self.extracted_flags = []

    @staticmethod
    def _parse_arguments():
        parser = argparse.ArgumentParser(description="Interactive Command Generator")
        parser.add_argument(
            "--results_dir",
            type=str,
            default="./results",
            help="Directory to save command results",
        )
        parser.add_argument(
            "--model_dir",
            type=str,
            default="./unified_models",
            help="Path to the model directory",
        )
        parser.add_argument(
            "--testing_mode",
            type=bool,
            default=False,
            help="Run vulnerability scans but do not attempt any exploits",
        )
        parser.add_argument(
            "--targets_list",
            type=str,
            default="targets.txt",
            help="lists of targets for autonomous testing",
        )
        parser.add_argument(
            "--autonomous_mode",
            type=bool,
            default=False,
            help="Flag to indicate autonomous mode",
        )
        parser.add_argument(
            "--attack_mode",
            type=str,
            default="stealth",
            help="Attack approach",
        )
        return parser.parse_args()

    def split(self, data):
        return data.split(": ", 1)

    def unique_commands_based_on_params(self, commands, ip_address):
        """
        Given a list of commands, returns all commands with unique sets of significant parameters.
        Excludes commands with nothing after the colon, commands with parameter '-p', and commands containing '-*'.
        Replaces [IP] with a provided IP address.
        """
        seen_params = set()
        unique_cmds = []

        for command in commands:
            # Extract the part of the command after the colon, if it exists
            parts = self.split(command)
            if len(parts) != 2 or not parts[1].strip():
                continue

            actual_command = parts[1].strip().rstrip(".").replace("[IP]", ip_address)

            try:
                # Split the command into tokens
                tokens = shlex.split(actual_command)
            except ValueError:
                print(f"Error splitting command: {actual_command}")
                logging.error(f"Error splitting command: {actual_command}")
                continue

            # Identify parameters in the tokens, excluding '-p'
            params = frozenset(
                token
                for token in tokens
                if (token.startswith("--") or token.startswith("-")) and token != "-p"
            )

            # Check if we've seen these parameters before
            if params not in seen_params:
                seen_params.add(params)
                if actual_command.startswith("nmap"):
                    timestamp = (
                        datetime.now()
                        .strftime("%I:%M:%S-%p-%Y-%m-%d")
                        .replace(" ", "-")
                    )
                    output_xml = f"results/nmap_output_{timestamp}.xml"
                    output_txt = f"results/nmap_output_{timestamp}.txt"
                    actual_command += f" -oX {output_xml} -oN {output_txt}"
                unique_cmds.append(actual_command)

        return unique_cmds

    def is_simple_port_scan(self, nmap_command):
        # Ensure the command is an nmap command
        if "nmap" not in nmap_command:
            return False

        # Check if script is in the command
        if "--script" in nmap_command:
            return False

        return True

    def autonomous_mode(self):
        url = ""

        def get_number_of_results(mode):
            return {"stealth": 1, "raid": 5, "war": 1e6}.get(mode, 1)

        def handle_command(command):
            if self.is_simple_port_scan(command) or command in command_history:
                cprint(f"Not running duplicate or useless command: {command}", "red")
                return
            cprint(f"Running command: {command}", "yellow")
            if not self.args.testing_mode:
                self.run_command_and_alert(command)
            match = re.search(r"^(.*?)-oX", command)
            if match:
                command = match.group(1)
            command_history.append(command)

        if not os.path.exists(self.args.targets_list):
            logging.error("The specified targets file does not exist, exiting...")
            cprint("The specified targets file does not exist, exiting...", "red")
            return

        command_history = []
        results = []

        with open(self.args.targets_list, "r") as file:
            for ip in file:
                ip = ip.strip()
                if not self.IP_PATTERN.match(ip):
                    cprint(f"Invalid IP: {ip} skipping it", "red")
                    continue

                cprint(
                    f"Running nmap vulnerability scans against {ip}, it may take several minutes please wait...",
                    "yellow",
                )
                timestamp = (
                    datetime.now().strftime("%I:%M:%S-%p-%Y-%m-%d").replace(" ", "-")
                )
                output_xml = f"results/nmap_output_{timestamp}.xml"
                output_txt = f"results/nmap_output_{timestamp}.txt"
                if self.args.testing_mode:
                    result = self.run_command_and_alert(
                        f"nmap -Pn --script=vuln {ip} -oX {output_xml} -oN {output_txt}"
                    )
                else:
                    result = self.run_command_and_alert(
                        f"nmap -Pn --script=vuln,exploit {ip} -oX {output_xml} -oN {output_txt}"
                    )
                results.append(result)

                cprint(f"nmap scanning completed for {ip}", "green")

        processed_data = self._parse_nmap_xml(output_xml)
        number_of_results = get_number_of_results(self.args.attack_mode)
        search_results = self.search_index(
            processed_data, self.return_path("indexdir_auto"), number_of_results
        )

        for data in processed_data:
            unique_commands = self.unique_commands_based_on_params(
                search_results, data["ip"]
            )
            for comm in tqdm(unique_commands, desc="Processing commands"):
                try:
                    ip = data["ip"]

                    handle_command(self.process_string(comm, [ip]))
                except Exception as e:
                    logging.error(f"Unable to run command, error: {e}")
                    cprint("Unable to run command, error", "red")

        cprint("Consulting models", "yellow")
        for model_name in self.model_names:
            if model_name in ["zap", "scribe"]:
                continue
            self._load_tokenizer_and_model(model_name)

            for data in processed_data:
                ip = data["ip"]
                for port, service in zip(data["ports"], data["services"]):
                    if model_name == "nuclei" and port not in ["80", "443"]:
                        continue
                    constructed_query = f"{service} {port} {ip}"
                    if port in ["80", "443"]:
                        url = f"https://{ip}" if port == "443" else f"http://{ip}"
                        constructed_query = (
                            f"run an automatic scan on {url} using the latest templates"
                        )
                    elif model_name == "crackmap":
                        constructed_query += " with a null session"

                    cprint(f"Constructed query: {constructed_query}", "green")
                    generated_text = self.generate_text(constructed_query.strip())
                    clean_up = self.process_string(
                        self.ensure_space_between_letter_and_number(generated_text),
                        [ip],
                        [url],
                    )
                    if model_name == "nmap":
                        clean_up = self.process_string(clean_up, [ip], [url], [port])
                    handle_command(clean_up)

    def select_mode(self):
        style = Style.from_dict(
            {
                "prompt": "cyan",
            }
        )

        action = (
            prompt(
                "\nYou can choose to load all models at once or on demand (for computers with less RAM/No GPU(s)) \n (a) load all (d) on demand? [a/d]: ",
                style=style,
            )
            .strip()
            .lower()
        )

        return action

    def print_farewell_message(self, width=30, height=10, density=0.5):
        # Calculate the position to print the farewell message
        farewell_msg = "Until our stars align again!!"
        start_x = (width - len(farewell_msg)) // 2
        start_y = height // 2

        star_colors = [
            "cyan",
            "magenta",
            "yellow",
            "blue",
        ]  # Nebula-themed colors for stars

        for y in range(height):
            for x in range(width):
                # Check if we are at the position to print the farewell message
                if y == start_y and start_x <= x < start_x + len(farewell_msg):
                    print(
                        colored(farewell_msg[x - start_x], "green", attrs=["bold"]),
                        end="",
                    )
                    continue

                if random.random() < density:
                    chosen_star = "*" if random.random() < 0.5 else "."
                    print(colored(chosen_star, random.choice(star_colors)), end="")
                else:
                    print(" ", end="")
            print()  # Move to the next line after each row

    def print_star_sky(self, width=30, height=10, density=0.5):
        # Calculate the position to print the welcome message
        welcome_msg = "Welcome to Nebula"
        start_x = (width - len(welcome_msg)) // 2
        start_y = height // 2

        star_colors = [
            "cyan",
            "magenta",
            "yellow",
            "blue",
        ]  # Nebula-themed colors for stars

        for y in range(height):
            for x in range(width):
                # Check if we are at the position to print the welcome message
                if y == start_y and start_x <= x < start_x + len(welcome_msg):
                    print(
                        colored(welcome_msg[x - start_x], "green", attrs=["bold"]),
                        end="",
                    )
                    continue

                if random.random() < density:
                    chosen_star = "*" if random.random() < 0.5 else "."
                    print(colored(chosen_star, random.choice(star_colors)), end="")
                else:
                    print(" ", end="")
            print()  # Move to the next line after each row

    def _load_flag_descriptions(self, file_path):
        """Load flag descriptions from a file and return them as a dictionary."""
        with open(file_path, "r") as f:
            lines = f.readlines()
            # Store the entire line as the value in the dictionary using flag as key
            return {
                line.split(":")[0].strip(): line.strip()
                for line in lines
                if ":" in line
            }

    def extract_and_match_flags(self, command):
        """Extract flags from the command and match them with descriptions."""
        flags = self.FLAG_PATTERN.findall(command)

        # Using set comprehension to ensure uniqueness
        matched_descriptions = list(
            {
                self.flag_descriptions[flag]
                for flag in flags
                if flag in self.flag_descriptions
            }
        )

        self.extracted_flags.extend(matched_descriptions)
        return matched_descriptions

    def return_path(self, path):
        if self.is_run_as_package():
            with resources.path("nebula", path) as correct_path:
                return str(correct_path)
        return path

    @staticmethod
    def _determine_s3_url():
        return (
            "https://nebula-models.s3.amazonaws.com/unified_models_no_zap.zip"
            if os.environ.get("IN_DOCKER")
            else "https://nebula-models.s3.amazonaws.com/unified_models.zip"
        )

    def _ensure_model_folder_exists(self):
        if not self.folder_exists_and_not_empty(self.args.model_dir):
            cprint(
                f"{self.args.model_dir} not found or is empty. Downloading and unzipping...",
                "yellow",
            )
            self.download_and_unzip(self.s3_url, f"{self.args.model_dir}.zip")
        else:
            cprint(
                f"found {self.args.model_dir}, to download new models remove {self.args.model_dir} or invoke nebula from a different directory",
                "green",
            )

    def _validate_model_dirs(self):
        model_dirs = [
            d
            for d in os.listdir(self.args.model_dir)
            if os.path.isdir(os.path.join(self.args.model_dir, d))
        ]
        if not model_dirs:
            raise ValueError("No model directories found in the specified directory.")

    def _ensure_results_directory_exists(self):
        if not os.path.exists(self.args.results_dir):
            os.makedirs(self.args.results_dir)

    def is_run_as_package(self):
        # Check if the script is within a 'site-packages' directory
        return "site-packages" in os.path.abspath(__file__)

    def folder_exists_and_not_empty(self, folder_path):
        # Check if the folder exists
        if not os.path.exists(folder_path) or not os.path.isdir(folder_path):
            return False
        # Check if the folder is empty
        if not os.listdir(folder_path):
            return False
        return True

    def download_and_unzip(self, url, output_name):
        try:
            # Download the file from the S3 bucket using wget with progress bar
            print("Downloading...")
            subprocess.run(
                ["wget", "--progress=bar:force:noscroll", url, "-O", output_name]
            )

            # Create the target directory if it doesn't exist
            target_dir = os.path.splitext(output_name)[0]
            os.makedirs(target_dir, exist_ok=True)

            # Unzip the downloaded file to the target directory
            print("\nUnzipping...")
            subprocess.run(["unzip", output_name, "-d", target_dir])

            os.remove(output_name)
        except subprocess.CalledProcessError as e:
            cprint(f"Error occurred: {e}", "red")
            logging.error(f"Error occurred: {e}")
        except Exception as e:
            cprint(f"Unexpected error: {e}", "red")
            logging.error(f"Unexpected error: {e}")

    def run_command_and_alert(self, text: str) -> None:
        """
        A function to run a command in the background, capture its output, and print it to the screen.
        """

        def execute_command(command: Union[str, List[str]]) -> Tuple[int, str, str]:
            """
            Executes the provided command and returns the returncode, stdout, and stderr.
            """
            try:
                process = subprocess.Popen(
                    command,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    text=True,
                )
                stdout, stderr = process.communicate()

                # Reset the terminal to a sane state after subprocess execution
                os.system("stty sane")

                return process.returncode, stdout, stderr
            except Exception as e:
                logging.error(f"Error while executing command {command}: {e}")
                return -1, "", str(e)

        if not self.args.autonomous_mode:
            cprint(
                "\nExecuting command, you can choose the view previous command option in the main menu to view the results when command execution has been completed",
                "yellow",
            )

        if isinstance(text, list):
            command_str = " ".join(text)
        else:
            command_str = text

        # Execute the command
        returncode, stdout, stderr = execute_command(command_str)
        truncated_cmd = command_str[:10].replace(" ", "_") + (
            "..." if len(command_str) > 15 else ""
        )

        timestamp = datetime.now().strftime("%I:%M:%S-%p-%Y-%m-%d").replace(" ", "-")
        file_name = f"{timestamp}_{truncated_cmd}"
        result_file_path = os.path.join(
            self.args.results_dir,
            f"result_{file_name}.txt",
        )
        try:
            ET.ElementTree(ET.fromstring(stdout))
            cprint("XML format detected, nothing to do", "green")
            return
        except Exception:

            # Conditions to decide if the results should be written to the file or not
            should_write_stderr = stderr and self.args.autonomous_mode is False
            if stdout.startswith("Starting Nmap"):
                return False
            if stderr.strip() or stdout.strip():
                with open(result_file_path, "w") as f:
                    if stdout.strip():
                        f.write(stdout)
                        return stdout
                    else:
                        if should_write_stderr and stderr.strip():
                            cprint("\nCommand Error Output:", "red")
                            cprint(stderr, "red")

                            f.write(stderr)
                            logging.error(
                                f"Command '{command_str}' failed with error:\n{stderr}"
                            )
                            cprint("\nhit the enter key to continue", "yellow")
                            return False

    @staticmethod
    def ensure_space_between_letter_and_number(s: str) -> str:
        try:
            if not isinstance(s, str):
                raise ValueError("The input must be a string.")

            # Regex operations
            # Match everything up to the first colon that's not immediately followed by a MAC address or an IP address
            s = re.sub(
                r"^.*?:\s(?!(?:[0-9a-fA-F]{2}:)+)(?!\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})",
                "",
                s,
            )

            s = re.sub(r"-(?=\d)", "", s)
            s = re.sub(r"(?<=[a-zA-Z])(?=\d)", " ", s)
            s = re.sub(r"(?<=\d)(?=[a-zA-Z])", " ", s)
            s = re.sub(r"\.$", "", s)

            return s.strip()

        except Exception as e:
            cprint(f"Error in ensure_space_between_letter_and_number: {e}", "red")
            logging.error(f"Error in ensure_space_between_letter_and_number: {e}")
            return s

    @staticmethod
    def _extract_ip(s: str) -> Optional[str]:
        """
        Extracts the first IP address found in the given string based on a pattern.
        If no IP address is found, it returns None.
        """
        try:
            ips = InteractiveGenerator.IP_PATTERN.findall(s)
            return ips if ips else None
        except Exception as e:
            logging.error(f"Error while extracting IP from string '{s[:30]}...': {e}")
            return None

    def process_string(
        self,
        s: str = "",
        replacement_ips: Optional[List[str]] = None,
        replacement_urls: Optional[List[str]] = None,
        port_arg: Optional[int] = None,
    ) -> str:

        # Handle the default values
        if replacement_ips is None:
            replacement_ips = []

        if replacement_urls is None:
            replacement_urls = []

        # Handle the default values
        if replacement_ips is None:
            replacement_ips = []

        if replacement_urls is None:
            replacement_urls = []
        """Replace the IP addresses and URLs in the given string with the respective replacements."""

        def get_local_ip() -> str:
            """Get local machine IP"""
            try:
                # This creates a new socket and connects to an external server's port.
                # We use Google's public DNS server for this example, but no data is actually sent.
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s.connect(("8.8.8.8", 80))
                ip = s.getsockname()[0]
                s.close()
                return ip
            except Exception:
                return "127.0.0.1"  # default to loopback, if unable to determine IP

        def get_random_port(above: int = 1000, max_retries: int = 100) -> int:
            """Get random port above the specified number that's not in use."""
            for _ in range(max_retries):
                port = random.randint(above, 65535)
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    try:
                        s.bind(("0.0.0.0", port))
                        return port
                    except socket.error:
                        continue
            raise ValueError("Unable to find an available port after multiple retries.")

        try:
            if not isinstance(s, str):
                raise ValueError("The input string must be a string.")
        except Exception as e:
            logging.error(f"Error in input string validation: {e}")

        # IP processing
        try:
            # Check if replacement_ips is a single IP and not a list
            if not isinstance(replacement_ips, list):
                replacement_ips = [replacement_ips]

            for ip in replacement_ips:
                if not re.match(self.IP_PATTERN, ip):
                    raise ValueError(f"One of the replacement IPs ({ip}) is not valid.")

            ip_addresses = re.findall(self.IP_PATTERN, s)
            if ip_addresses:
                for i, ip in enumerate(ip_addresses):
                    if i < len(replacement_ips):
                        s = s.replace(ip, replacement_ips[i])
        except Exception as e:
            logging.error(f"Error in IP processing: {e}")

        # URL processing
        try:
            if not isinstance(replacement_urls, list):
                replacement_urls = [replacement_urls]

            for url in replacement_urls:
                if not re.match(self.URL_PATTERN_VALIDATION, url):
                    raise ValueError(
                        f"One of the replacement URLs ({url}) is not valid."
                    )

            urls = self.extract_urls(s)
            for i, url in enumerate(urls):
                if i < len(replacement_urls):
                    s = s.replace(url, replacement_urls[i])
        except Exception as e:
            logging.error(f"Error in URL processing: {url}{e}")

        # Replace placeholders
        if replacement_ips and len(replacement_ips) > 0:
            primary_ip = replacement_ips[0]
            s = s.replace("{{ RHOSTS }}", primary_ip)
            if port_arg:
                s = s.replace("{{ RPORT }}", str(port_arg))
            s = s.replace("{{ LHOST }}", get_local_ip())
            s = s.replace("{{ LPORT }}", str(get_random_port()))
        timestamp = datetime.now().strftime("%I:%M:%S-%p-%Y-%m-%d").replace(" ", "-")
        if s.strip().startswith("nmap"):
            output_xml = f"results/nmap_output_{timestamp}.xml"
            output_txt = f"results/nmap_output_{timestamp}.txt"
            s += f" -oX {output_xml} -oN {output_txt}"

        return s

    def extract_urls(self, s: str) -> list:
        """Extract URLs from the given string.

        Args:
            s (str): The input string.

        Returns:
            list: List of URLs found in the string.
        """
        url_pattern = r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
        return re.findall(url_pattern, s)

    def generate_text(self, prompt_text: str, max_length: int = 300000) -> str:
        """
        Generate text using the current model based on the provided prompt.

        Parameters:
        - prompt_text: Initial text to prompt the model for generation.
        - max_length: Maximum length for the generated text.

        Returns:
        - The generated text as a string.
        """
        try:
            if len(prompt_text) > self.current_model.config.n_ctx:
                logging.warning("Prompt too long! Truncating...")
                cprint("Prompt too long! Truncating...", "red")
                prompt_text = prompt_text[: self.current_model.config.n_ctx]

            encoding = self.current_tokenizer.encode_plus(
                prompt_text,
                return_tensors="pt",
                add_special_tokens=True,
                max_length=max_length,
                pad_to_max_length=False,
                return_attention_mask=True,
                truncation=True,
            )

            input_ids = encoding["input_ids"].to(self.device)
            attention_mask = encoding["attention_mask"].to(self.device)
            temp = 0.1
            if self.current_model == "scribe":
                temp = 0.5
            with tqdm(total=max_length, desc="Generating text", position=0) as pbar:
                output = self.current_model.generate(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    max_length=max_length,
                    num_return_sequences=1,
                    do_sample=True,
                    top_k=40,
                    top_p=0.95,
                    temperature=temp,
                    repetition_penalty=1.0,
                )
                pbar.update(len(output[0]))

            generated_text = self.current_tokenizer.decode(
                output[0], skip_special_tokens=True
            )

            return generated_text

        except Exception as e:
            logging.error(
                f"Error during text generation with prompt '{prompt_text[:30]}...': {e}"
            )
            return ""  # Returning an empty string as a fallback. Adjust based on your requirements.

    def start_interactive_mode(self) -> None:
        """
        Start an interactive session for the user to input prompts and receive generated text.

        In the interactive session:
        - The user is prompted to enter text.
        - The model generates text based on the user's input.
        - The generated text is displayed and handled accordingly.
        - The session can be exited by typing 'exit' or through keyboard interrupts.
        """

        while True:
            try:
                prompt = self.get_user_prompt()

                if not prompt:
                    print("Prompt is empty. Please enter a valid prompt.")
                    continue

                if prompt.lower() == "exit":
                    break

                generated_text = self.generate_and_display_text(prompt)
                self.handle_generated_text(generated_text)

            except (EOFError, KeyboardInterrupt):  # Combined the two exception types
                self.print_farewell_message()
                break
            except Exception as e:  # Handle unexpected errors
                if isinstance(prompt, bool):
                    logging.error(
                        f"Error during interactive session with prompt '{prompt}...': {e}"
                    )
                elif isinstance(prompt, list):
                    logging.error("prompt value is:%s", prompt)
                    logging.error(
                        f"Error during interactive session with prompt '{prompt[:30]}...': {e}"
                    )

    def display_command_list(self, file_extension=None) -> None:
        """
        Display a list of previously saved result filenames truncated to self.max_truncate_length.
        """
        cprint("\nPrevious Results:", "cyan")

        # Fetching filenames from the results directory
        all_files = os.listdir(self.args.results_dir)

        # If a file extension is specified, filter the files by that extension
        if file_extension:
            filenames = sorted(
                [file for file in all_files if file.endswith(file_extension)]
            )
        else:
            filenames = sorted(all_files)
        if not filenames:
            cprint("no results found", "red")
            return
        for idx, filename in enumerate(filenames):
            truncated_filename = (
                (filename[: self.max_truncate_length] + "...")
                if len(filename) > self.max_truncate_length
                else filename
            )
            cprint(f"{idx+1}. {truncated_filename}", "yellow")

    def get_user_prompt(self) -> str:
        """
        Interactively prompt the user for an action or command input.

        Returns:
            str: The user's command input.
        """
        style = Style.from_dict({"prompt": "cyan", "error": "red"})

        while True:
            try:
                if self.command_running:
                    action = (
                        prompt(
                            "\nA command is currently running. Do you want to (w) wait for it to complete, or (c) continue without waiting? [w/c]: ",
                            style=style,
                        )
                        .strip()
                        .lower()
                    )

                    if action == "w":
                        cprint("Waiting for the command to complete...", "yellow")
                        while self.command_running:
                            time.sleep(1)
                        cprint(
                            "Command completed!, you can view the result using the 'view previous results' option on the main menu",
                            "green",
                        )
                        continue

                action = (
                    prompt(
                        "\nDo you want to (c) enter a new command, (v) view previous results, (pr) process previous nmap results,\n(m) select a model, (s) to search by keywords or (q) quit? [c/v/m/s/q]: ",
                        style=style,
                    )
                    .strip()
                    .lower()
                )

                if action == "q":
                    self.print_farewell_message()
                    exit(0)
                elif action == "v":
                    return self._view_previous_results(".txt")
                elif action == "pr":
                    return self._process_previous_nmap_results(".xml")
                elif action == "c":
                    return self._input_command_without_model_selection()
                elif action == "m":
                    self._select_model()
                    continue
                elif action == "s":
                    return self.user_search_interface()
                else:
                    cprint("Invalid option. Please choose a valid option.", "red")

            except Exception as e:
                logging.error(f"Error during user prompt: {e}")
                cprint("An unexpected error occurred. Please try again.", "red")

    def contains_pytorch_model(self, directory: str) -> bool:
        """
        Check if the given directory contains a .bin file (assumed to be a PyTorch model file).
        """
        if not os.path.isdir(directory):  # Ensure it's a directory
            return False

        for filename in os.listdir(directory):
            if filename.endswith(".bin"):
                return True
        return False

    def get_model_names(self):
        self.model_names = [
            model_name
            for model_name in os.listdir(self.args.model_dir)
            if self.contains_pytorch_model(
                os.path.join(self.args.model_dir, model_name)
            )
        ]
        return self.model_names

    def _select_model(self) -> str:
        """
        Interactively allows the user to select a model from the available models.

        Returns:
            str: The name of the selected model.
        """
        selected_model_name = None

        session = PromptSession()

        while True:
            print_formatted_text(
                "Available models:", style=Style.from_dict({"": "yellow"})
            )
            for idx, model_name in enumerate(self.model_names, 1):
                print_formatted_text(
                    f"{idx}. {model_name}", style=Style.from_dict({"": "yellow"})
                )

            choice = session.prompt(
                "Select a model by entering its number: ",
                style=Style.from_dict({"": "cyan"}),
            )

            if choice.isdigit() and 1 <= int(choice) <= len(self.model_names):
                selected_model_name = self.model_names[int(choice) - 1]
                break
            else:
                print_formatted_text(
                    "Invalid choice. Please enter a valid number.",
                    style=Style.from_dict({"": "red"}),
                )

        # Handling the single model mode
        if self.single_model_mode:
            self._load_tokenizer_and_model(selected_model_name)
        else:
            self.current_model = self.models[selected_model_name]
            self.current_tokenizer = self.tokenizers[selected_model_name]

        # Handling model-specific behavior
        if selected_model_name == "nmap":
            self.flag_file = self.return_path("nmap_flags")
        elif selected_model_name == "crackmap":
            self.flag_file = self.return_path("crackmap_flags")
        elif selected_model_name == "nuclei":
            self.flag_file = self.return_path("nuclei_flags")
        elif selected_model_name == "zap":
            self.flag_file = self.return_path("zap_flags")

        self.flag_descriptions = self._load_flag_descriptions(self.flag_file)

        Style.from_dict({"message": "bg:#ff0066 #ffff00"})

        print_formatted_text(
            f"You've selected the {selected_model_name} model!",
            style=Style.from_dict({"": "yellow"}),
        )

        return selected_model_name

    def unload_model(self):
        # 1. Explicitly delete models and tokenizers
        if hasattr(self, "current_model") and self.current_model:
            del self.current_model
            self.current_model = None

        if hasattr(self, "current_tokenizer") and self.current_tokenizer:
            del self.current_tokenizer
            self.current_tokenizer = None

        # 2. Release GPU Memory
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        # 3. Clear any cached states in model's `from_pretrained` methods
        # This is specific to transformers library and can be useful for some models
        if hasattr(GPT2LMHeadModel, "clear_cache"):
            GPT2LMHeadModel.clear_cache()

        # 4. Force garbage collection
        gc.collect()

    def _load_tokenizer_and_model(self, model_name: Optional[str] = None):
        # Clear current models and tokenizers if a specific model is to be loaded
        first_loaded = None
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )  # Set the device once

        # List all available model directories or just the specified one
        model_folders = [model_name] if model_name else os.listdir(self.args.model_dir)

        # Iterating over the list of model directories
        for model_folder in model_folders:
            full_path = os.path.join(self.args.model_dir, model_folder)

            # Check if it's a directory to avoid loading from non-directory paths
            if os.path.isdir(full_path):
                try:
                    # Load tokenizer
                    if self.current_model and self.single_model_mode:
                        self.unload_model()
                    cprint(
                        f"Loading tokenizer for {model_folder}...",
                        "yellow",
                        end="",
                        flush=True,
                    )
                    self.current_tokenizer = GPT2Tokenizer.from_pretrained(full_path)
                    cprint(" Done!", "green")

                    # Load model
                    cprint(
                        f"Loading model for {model_folder}...",
                        "yellow",
                        end="",
                        flush=True,
                    )
                    self.current_model = GPT2LMHeadModel.from_pretrained(full_path)
                    self.current_model.eval()
                    self.current_model.to(self.device)
                    cprint(" Done!", "green")

                    # Add the successfully loaded model and tokenizer to their respective dictionaries
                    if not self.single_model_mode:
                        self.tokenizers[model_folder] = self.current_tokenizer
                        self.models[model_folder] = self.current_model

                    # Only set the 'first_loaded' once
                    if first_loaded is None:
                        first_loaded = model_folder

                except Exception as e:
                    cprint(
                        f"Failed to load model/tokenizer from {model_folder}: {e}",
                        "red",
                    )
                    logging.error(
                        f"Failed to load model/tokenizer from {model_folder}: {e}"
                    )

        # If not in single_model_mode or no model was specified, set the first loaded model as the active model
        if (
            not hasattr(self, "single_model_mode") or not self.single_model_mode
        ) and first_loaded:
            self.current_tokenizer = self.tokenizers[first_loaded]
            self.current_model = self.models[first_loaded]
            cprint(f"\nThe current model in use is: {first_loaded}\n", "blue")

        # Memory usage information
        try:
            process = psutil.Process(os.getpid())
            cprint(
                f"CPU memory usage: {process.memory_info().rss / (1024**2):.2f} MB",
                "green",
            )
        except Exception as e:
            cprint(f"An error occurred while fetching CPU memory info: {e}", "red")
            logging.error(f"An error occurred while fetching CPU memory info: {e}")

        try:
            pynvml.nvmlInit()
            handle = pynvml.nvmlDeviceGetHandleByIndex(0)
            info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            cprint(f"Used GPU memory: {info.used / (1024**2):.2f} MB", "green")

        except Exception as e:
            cprint(f"An error occurred while fetching GPU memory info: {e}", "red")
            logging.error(f"An error occurred while fetching GPU memory info: {e}")

        return first_loaded

    def _input_command_without_model_selection(self) -> str:
        """Internal method to get a command input from the user without model selection."""

        # Styling definitions
        style = Style.from_dict({"prompt": "cyan", "error": "red", "message": "green"})

        # Initialize the spell checker
        spell = SpellChecker()

        while True:  # Keep prompting until valid input or 'q' is entered
            # Get the actual user input for the model to generate
            user_input = prompt(
                "\nEnter a prompt (or 'b' to return to the main menu): ", style=style
            ).strip()

            # If the user wants to return to the previous menu
            if user_input.lower() == "b":
                # Here, you can define what should be done to return to the previous menu.
                # For this example, we'll just return an empty string.
                return True

            # If input is empty, prompt again
            if not user_input:
                cprint("Please provide a prompt or enter 'q' to return.", "red")
                continue

            # Split the user input into words
            words = user_input.split()

            # Correct words that are not in the dictionary
            corrections_made = False  # Flag to check if any corrections were accepted
            for index, word in enumerate(words):
                # Check if the word is a number or contains digits; if yes, then continue to the next iteration
                if word.isdigit() or any(char.isdigit() for char in word):
                    continue

                # Check if word is a URL; if yes, continue to the next iteration
                if re.match(
                    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
                    word,
                ):
                    continue

                # If the word is not in the dictionary
                misspelled = spell.unknown([word])
                if misspelled:
                    # Get the most likely correct spelling for the word
                    suggestion = spell.correction(word)

                    # If no suggestions are found, continue to the next iteration
                    if not suggestion:
                        continue

                    # Display the suggestion to the user and ask for their choice
                    try:
                        choice = prompt(
                            f"Did you mean '{suggestion}' instead of '{word}'? (Y/n): ",
                            style=style,
                            validator=WordValidator(["y", "n", ""]),
                            validate_while_typing=False,
                        ).lower()
                    except Exception as e:
                        cprint(f"Error occurred: {e}", "red")
                        logging.error(f"Error occurred: {e}")

                    # If the user accepts the suggestion, replace the word in the list
                    if choice == "y" or choice == "":
                        words[index] = suggestion
                        corrections_made = True  # Update the flag

            if corrections_made:
                corrected_input = " ".join([str(word) for word in words])
            else:
                corrected_input = user_input

            if user_input.lower() in ["quit", "exit"]:
                self.print_farewell_message()
                exit(0)

            return corrected_input

    def search_index(
        self, query_list: Union[list, str], indexdir: str, max_results: int = 10
    ) -> list:
        try:
            ix = open_dir(indexdir)
        except Exception as e:
            logging.error(f"Error occurred while opening index directory: {e}")
            return []

        results_dict = {}
        searched_items = set()

        if isinstance(query_list, str):
            query_list = [{"services": [query_list]}]

        with ix.searcher() as searcher:
            # Multifield Search
            query_parser = MultifieldParser(
                ["content"], schema=ix.schema, group=OrGroup
            )

            for query in query_list:
                cves = query.get("cves", [])
                service_names = query.get("services", [])

                # Search for CVEs first if they exist
                for cve in cves:
                    if cve.lower() in searched_items:
                        continue

                    searched_items.add(cve.lower())
                    parsed_query = query_parser.parse(cve.lower())
                    cve_results = []

                    try:
                        results = searcher.search(parsed_query, limit=max_results)
                        for res in results:
                            content = res["content"]
                            lines = content.splitlines()
                            for line in lines:
                                if ":" in line:
                                    # Get the part before the colon
                                    before_colon = line.split(":")[0]
                                    if cve.lower() in before_colon.lower():
                                        cve_results.append(line.strip())
                                        if len(cve_results) >= max_results:
                                            break

                        if cve_results:
                            results_dict[cve] = cve_results

                    except Exception as e:
                        logging.error(
                            f"Error occurred while searching for CVE {cve}: {e}"
                        )

                # If services are specified, search for them
                for s in service_names:
                    if s.lower() in searched_items:
                        continue

                    searched_items.add(s.lower())
                    parsed_query = query_parser.parse(s.lower())
                    service_results = []

                    try:
                        results = searcher.search(parsed_query, limit=max_results)
                        for res in results:
                            content = res["content"]
                            lines = content.splitlines()
                            for line in lines:
                                if ":" in line:
                                    # Get the part before the colon
                                    before_colon = line.split(":")[0]
                                    if s.lower() in before_colon.lower():
                                        service_results.append(line.strip())
                                        if len(service_results) >= max_results:
                                            break

                        if service_results:
                            results_dict[f"Service {s}"] = service_results

                    except Exception as e:
                        logging.error(
                            f"Error occurred while searching for service {s}: {e}"
                        )

        formatted_results = []
        for key, values in results_dict.items():
            formatted_results.append(key + ":")
            formatted_results.extend(values)

        return formatted_results

    def _parse_nmap_xml(self, xml_file):
        if os.path.isfile(xml_file):
            tree = ET.parse(xml_file)
        else:
            tree = ET.ElementTree(ET.fromstring(xml_file))
        root = tree.getroot()

        parsed_results = []
        cve_matches = set()  # Use a set to avoid duplicate CVEs

        # Extract CVEs from the entire XML once
        for elem in root.iter():  # Search within the entire XML tree
            # Check element tag for CVE
            tag_cves = self.CVE_PATTERN.findall(elem.tag)
            cve_matches.update(tag_cves)

            # Check element text for CVE
            if elem.text:
                text_cves = self.CVE_PATTERN.findall(elem.text)
                cve_matches.update(text_cves)

            # Check all attributes for CVEs
            for attrib_name, attrib_value in elem.attrib.items():
                attrib_name_cves = self.CVE_PATTERN.findall(attrib_name)
                cve_matches.update(attrib_name_cves)

                attrib_value_cves = self.CVE_PATTERN.findall(attrib_value)
                cve_matches.update(attrib_value_cves)

        for attrib_value_cve in cve_matches:
            cprint(f"CVE(s) found: {attrib_value_cve}", "red")
        for host in root.findall("host"):
            try:
                device_name = host.find("hostnames/hostname").attrib.get(
                    "name", "Unknown"
                )
            except AttributeError:
                device_name = "Unknown"

            try:
                ip_address = host.find("address").attrib.get("addr", "Unknown")
            except AttributeError:
                ip_address = "Unknown"

            ports = []
            services = []

            for port in host.findall("ports/port"):
                try:
                    port_id = port.attrib.get("portid")
                    port_state = port.find("state").attrib.get("state")
                    service_name = port.find("service").attrib.get("name")

                    if port_state == "open" and port_id not in ports:
                        ports.append(port_id)
                        if service_name == "domain":
                            services.append("dns")
                        else:
                            services.append(service_name)
                except AttributeError:
                    continue

            parsed_results.append(
                {
                    "hostname": device_name,
                    "ip": ip_address,
                    "ports": ports,
                    "services": services,
                    "cves": list(
                        cve_matches
                    ),  # Convert the set to a list before adding
                }
            )

        return parsed_results

    def colored_output(self, text):
        """Color everything before the first ':' in red and the rest in yellow."""

        # Split the text at the first ':'
        parts = re.split("(:)", text, maxsplit=1)

        # Color everything before the first ':' in red
        if len(parts) > 1:
            colored_text = (
                f"\033[91m{parts[0]}\033[0m\033[93m{parts[1]}{parts[2]}\033[0m"
            )
        else:
            colored_text = f"\033[93m{text}\033[0m"

        return colored_text

    def get_input_with_validation(
        self, prompt_text: str, valid_fn: Optional[callable] = None
    ) -> str:
        style = Style.from_dict({"prompt": "cyan", "error": "red", "message": "green"})

        validator = None
        if valid_fn:
            validator = FunctionValidator(valid_fn)

        while True:
            try:
                user_input = prompt(
                    prompt_text,
                    style=style,
                    validator=validator,
                    validate_while_typing=False,
                )
                return user_input.lower().strip()
            except ValidationError:
                continue

    def is_valid_command_choice(self, choice: str, max_choice: int) -> bool:
        if choice in ["back", "b"]:
            return True
        if choice.isdigit() and 0 < int(choice) <= max_choice:
            return True
        return False

    def is_valid_results_choice(self, choice: str) -> bool:
        return choice.isdigit() and int(choice) > 0

    def get_input_with_default(self, message, default_text=None):
        style = Style.from_dict({"prompt": "cyan", "info": "cyan"})
        history = InMemoryHistory()
        if default_text:
            user_input = prompt(
                message, default=default_text, history=history, style=style
            )
        else:
            user_input = prompt(message, history=history, style=style)
        return user_input

    def _process_previous_nmap_results(self, file_extension=None) -> bool:
        """Process the results of a previously executed nmap command."""

        # Display the command list based on the optional file extension filter
        self.display_command_list(file_extension=file_extension)

        cmd = self.get_input_with_default(
            "Enter the number of the command you'd like to process results for (or type 'back' or 'b' to return): "
        )

        if cmd in ["back", "b"]:
            return True

        number_of_results = self.get_input_with_default(
            "Enter the number of the processed results you would like to receive (or type 'back' or 'b' to return, 'all' or 'a' for all results): "
        )

        if number_of_results in ["back", "b"]:
            return True

        return self._display_and_select_results(cmd, number_of_results, file_extension)

    def _display_and_select_results(
        self, cmd, number_of_results, file_extension=None
    ) -> bool:
        """Display the nmap results and prompt user for a result selection."""

        cmd_num = int(cmd)

        # Fetching filenames from the results directory based on the optional file extension filter
        all_files = os.listdir(self.args.results_dir)

        if file_extension:
            filenames = sorted(
                [file for file in all_files if file.endswith(file_extension)]
            )
        else:
            filenames = sorted(all_files)

        # Ensure the selected cmd_num is within range
        if 0 <= cmd_num <= len(filenames):
            selected_filename = filenames[cmd_num - 1]
            file_path = os.path.join(self.args.results_dir, selected_filename)
        else:
            print(f"Invalid command number: {cmd_num}. Please choose a valid command.")
            return True

        services = self._parse_nmap_xml(file_path)
        max_results_value = (
            int(number_of_results)
            if number_of_results not in ["all", "a"]
            else MAX_RESULTS_DEFAULT
        )
        results = self.search_index(services, self.index_dir, max_results_value)

        if not results:
            print("No results found.")
            return True

        self._display_search_results(results)
        self._select_and_run_command(results)

        return True

    def _display_search_results(self, results):
        """Display the search results."""

        idx = 1
        for line in results:
            is_service = line.strip().lower().startswith("service")
            is_cve = (
                line.strip().lower().startswith("cve")
                and ":" in line
                and line.split(":", 1)[1].strip()
            )

            if line.endswith(":"):
                print(colored(line, "yellow"))
            elif is_service:
                print(colored(line, "yellow"))  # Retain the yellow color for services
            elif is_cve:
                prefix, suffix = line.split(":", 1)  # Split the CVE line at the colon
                print(
                    colored(f"{idx}. {prefix}:", "red") + colored(f" {suffix}", "blue")
                )
                idx += 1
            else:
                prefix, suffix = line.split(":", 1) if ":" in line else (line, "")
                print(
                    colored(f"{idx}. {prefix}:", "white")
                    + colored(f" {suffix}", "green")
                )
                idx += 1

    def _select_and_run_command(self, results):
        """Prompt user to select a search result and then run a command based on that result."""

        selection = self.get_input_with_default(
            "\nSelect a result number to modify and run the associated command or any other key to continue without running a command: "
        )

        # Filtered results for selection
        selectable_results = [
            line
            for line in results
            if not (
                line.strip().lower().startswith("service")
                or (
                    line.strip().lower().startswith("cve")
                    and ":" in line
                    and not line.split(":", 1)[1].strip()
                )
            )
        ]

        try:
            selection = int(selection)
            if 1 <= selection <= len(selectable_results):
                content_after_colon = (
                    selectable_results[selection - 1].split(":", 1)[1].strip()
                    if ":" in selectable_results[selection - 1]
                    else selectable_results[selection - 1]
                )

                modified_content = self.get_input_with_default(
                    "\nEnter the modified content (or press enter to keep it unchanged): ",
                    content_after_colon,
                )

                if modified_content:
                    content_after_colon = modified_content

                self.run_command(content_after_colon)
            else:
                cprint("Invalid selection!", "red")
        except ValueError:
            pass

    def nmap_xml_to_plain_text(self, tree):
        # Parse the XML file

        root = tree.getroot()

        # Loop through each host in the XML
        for host in root.findall("host"):
            # Get IP address
            ip_address = host.find("address").get("addr")
            print(f"Host: {ip_address}")

            # Get hostnames
            hostnames = host.find("hostnames")
            for name in hostnames.findall("hostname"):
                print(f"Hostname: {name.get('name')}")

            # Get ports information
            ports = host.find("ports")
            for port in ports.findall("port"):
                port_id = port.get("portid")
                protocol = port.get("protocol")
                state = port.find("state").get("state")
                service = port.find("service").get("name")
                product = port.find("service").get("product", "")

                cprint(
                    f"Port: {port_id}/{protocol}, State: {state}, Service: {service} {product}",
                    "cyan",
                )

        print("---------")

    def _view_previous_results(self, file_extension=None) -> bool:
        """Internal method to display previous results and loop back to the main prompt."""

        # Fetch filenames from the results directory
        all_files = os.listdir(self.args.results_dir)

        # If a file extension is specified, filter the files by that extension
        if file_extension:
            filenames = sorted(
                [file for file in all_files if file.endswith(file_extension)]
            )
        else:
            filenames = sorted(all_files)

        # Check if there are no previous results
        if not filenames:
            cprint("No previous results available.", "red")
            return True  # Return to the main loop

        style = Style.from_dict({"prompt": "cyan"})

        while True:  # Keep looping until user decides to go back
            self.display_command_list(file_extension)

            cmd = prompt(
                "Enter the number of the result you'd like to view (or type 'back' or 'b' to return): ",
                style=style,
            )

            if cmd.lower().strip() in ["back", "b"]:
                return True  # Indicate that the user wants to return to the main loop

            try:
                cmd_num = int(cmd)
                if 0 < cmd_num <= len(filenames):
                    file_path = os.path.join(
                        self.args.results_dir, filenames[cmd_num - 1]
                    )

                    with open(file_path, "r") as f:
                        result = f.read()
                        cprint(f"\nResults for command #{cmd_num}:", "cyan")
                        cprint(result, "magenta")
                else:
                    cprint(
                        f"Invalid number. Please choose between 1 and {len(filenames)}.",
                        "red",
                    )
            except ValueError:
                cprint("Please enter a valid number or type 'back'.", "red")

    def generate_and_display_text(self, prompt: str) -> str:
        """Generate and display text based on a given prompt.

        Args:
            prompt (str): The input prompt for text generation.

        Returns:
            str: The cleaned-up generated text.
        """
        try:
            generated_text = self.generate_text(prompt)
            prompt_ip = self._extract_ip(prompt)
            urls = self.extract_urls(prompt)
            first_clean_up = self.ensure_space_between_letter_and_number(generated_text)
            second_clean_up = self.process_string(first_clean_up, prompt_ip, urls)
            if self.args.autonomous_mode is False:
                cprint("showing flags", "red")
                try:
                    help = self.extract_and_match_flags(second_clean_up)
                    if help:
                        cprint(
                            "Verify that the flags used in the command matches your intent:",
                            "magenta",
                        )
                        for h in help:
                            cprint("\n" + h, "red")
                except Exception as e:
                    logging.error(f"unable to extract, error: {e}")

            cprint("\nGenerated Command:", "cyan")
            cprint(second_clean_up, "magenta")
            cprint("-" * 50, "blue")

            return second_clean_up
        except Exception as e:
            logging.error(f"Error during text generation and display: {e}")
            cprint(
                "An unexpected error occurred during text generation and display. Please try again or check logs.",
                "red",
            )
            return False  # This returns the original prompt as a fallback. Adjust as needed.

    def run_command(self, text: str) -> None:
        """
        A function to run a command in the background based on the generated text.
        """

        def threaded_function():
            self.run_command_and_alert(text)
            self.command_running = (
                False  # Set the flag to False once the command is done executing.
            )

        # Before starting the thread, set the command_running flag to True.
        self.command_running = True
        time.sleep(
            0.1
        )  # Introduce a short delay to give main thread some breathing room
        thread = threading.Thread(target=threaded_function)
        thread.start()

        # Inform user that command has started
        print("\nThe operation has been initiated.")

        return

    def user_search_interface(self):
        """Provide a user interface for searching."""
        suggestions = self.get_suggestions()
        protocol_completer = WordCompleter(suggestions, ignore_case=True)
        history = InMemoryHistory()

        while True:
            query_str = self.get_query_input(
                completer=protocol_completer, history=history
            )
            if not query_str:
                self.display_message(
                    "Please enter a query or 'b' to return to the main menu", "red"
                )
                continue
            if query_str.lower() == "b":
                break

            num_results = self.get_num_results_input(history=history)
            results = self.get_search_results(query_str, num_results)

            if not results:
                self.display_message("No results found.", "red")
                return True

            self.display_results(results, history)
        return True

    def get_suggestions(self):
        suggestions_file = self.return_path("suggestions")
        with open(suggestions_file, "r") as file:
            return [line.strip() for line in file if line.strip()]

    def display_message(self, message, color):
        cprint(message, color)

    def get_query_input(self, completer, history):
        return prompt(
            ANSI(
                colored(
                    "\nEnter your query or 'b' to return to the main menu: ", "blue"
                )
            ),
            completer=completer,
            history=history,
        )

    def get_num_results_input(self, history):
        return prompt(
            ANSI(
                colored(
                    "\nHow many results per category should i display? (Enter a number, (a) or (all) for all results, or press enter for default:",
                    "blue",
                )
            ),
            history=history,
        )

    def get_search_results(self, query_str, num_results):
        if num_results.lower() in ["a", "all"]:
            return self.search_index(
                query_str, self.index_dir, max_results=float("inf")
            )
        if not num_results:
            return self.search_index(query_str, self.index_dir)
        try:
            num = int(num_results)
            return self.search_index(query_str, self.index_dir, max_results=num)
        except ValueError:
            self.display_message("Invalid input. Returning to previous menu.", "red")
            return []

    def display_results(self, results, history):
        service_lines = [line for line in results if line.startswith("Service ")]
        other_lines = [line for line in results if not line.startswith("Service ")]

        for line in service_lines:
            print(colored(line, "green"))

        for idx, line in enumerate(other_lines, 1):
            print(f"{idx}. {self.colored_output(line)}")

        self.handle_result_selection_and_modification(other_lines, history)

    def handle_result_selection_and_modification(self, other_lines, history):
        while True:
            try:
                selection = int(
                    prompt(
                        ANSI(
                            colored(
                                "\nSelect a result number to modify or any other key to continue: ",
                                "blue",
                            )
                        ),
                        history=history,
                    )
                )
                if 1 <= selection <= len(other_lines):
                    self.modify_and_run_command(other_lines[selection - 1])
                    break
                else:
                    self.display_message("Invalid selection!", "red")
            except ValueError:
                self.display_message(
                    "Invalid input! Returning to previous menu.", "red"
                )
                break

    def modify_and_run_command(self, selected_line):
        content_after_colon = (
            selected_line.split(":", 1)[1].strip()
            if ":" in selected_line
            else selected_line
        )
        modified_content = prompt(
            ANSI(
                colored(
                    "\nEnter the modified content or press enter to keep it unchanged: ",
                    "blue",
                )
            ),
            default=content_after_colon,
        )
        if modified_content:
            content_after_colon = modified_content
        self.run_command(content_after_colon)

    def format_service_name(self, line):
        """Format the service name properly."""

        # Split the line into hostname and service
        parts = line.split(":", 1)
        if len(parts) != 2:
            return line  # Return the original line if it doesn't contain a service name

        hostname, service = parts

        # Format the service name based on the desired format.
        # Example: If the service name is "domain", replace it with "dns"
        if service == "domain":
            service = "dns"

        # Combine the formatted service name with the hostname
        return f"{hostname}:{service}"

    def get_modified_command(self, text: str) -> str:
        """Prompt the user to modify and return a command.

        Args:
            text (str): The proposed command.

        Returns:
            str: The modified command.
        """
        history = InMemoryHistory()
        style = Style.from_dict(
            {
                "prompt": "yellow",
            }
        )
        try:
            print_formatted_text(
                ANSI(colored("\nThe current command is:", "cyan")), end=""
            )
            print_formatted_text(ANSI(colored(text, "magenta")))

            # Prompt the user to modify the command with a default value
            modified_text = prompt(
                "\nModify the command as needed and press Enter: ",
                default=text,
                history=history,
                style=style,
            )

            return modified_text.strip()

        except Exception as e:
            logging.error(f"Error during modified command prompt: {e}")
            print_formatted_text(
                ANSI(
                    colored(
                        "An unexpected error occurred while modifying the command. Please try again or check logs.",
                        "red",
                    )
                )
            )
            return text  # Return the original text as a fallback

    def handle_generated_text(self, text):
        """Handle the generated text based on user's choice or predefined actions.

        Args:
            text (str): The generated text or command.
        """
        if not text:
            cprint("No text generated", "red")
            return
        try:
            # If always_apply_action is set, simply run the command
            if self.always_apply_action:
                self.run_command(text)
                return

            # Ask the user for their action choice
            action_choice = self.get_action_choice()

            # If 'always' is chosen, set the flag and run the command
            if action_choice in ["always", "a"]:
                self.always_apply_action = True
                self.run_command(text)
            # If 'yes' is chosen, let the user modify the command and then run it
            elif action_choice in ["yes", "y"]:
                modified_command = self.get_modified_command(text)
                self.run_command(modified_command if modified_command else text)
        except Exception as e:
            logging.error(f"Error handling generated text: {e}")
            cprint(
                "An error occurred while handling the generated text. Please try again or check logs.",
                "red",
            )

    def get_action_choice(self) -> str:
        style = Style.from_dict(
            {"prompt": "yellow", "options": "green", "error": "red", "message": "blue"}
        )

        if self.current_model_name != "scribe":
            action_choice_prompt = (
                "Do you want to run a command based on the generated text?"
            )
            options = "(y/n/a) (yes/no/always):"

            while True:
                try:
                    action_choice = prompt(
                        f"{action_choice_prompt} {options}",
                        style=style,
                        validator=ActionChoiceValidator(),
                        validate_while_typing=False,
                    )
                    return action_choice.lower().strip()
                except ValidationError:
                    continue
                except Exception as e:
                    logging.error(f"Error during action choice prompt: {e}")
                    print_formatted_text(
                        "An unexpected error occurred during action choice. Please try again or check logs.",
                        style="error",
                    )
        else:
            return


def main_func():
    generator = InteractiveGenerator()
    generator.start_interactive_mode()


if __name__ == "__main__":
    main_func()
